;; 
;;  uim-candidate.el
;;
;;  Copyright (c) 2005, Konosuke Watanabe <nosuke@csc.ne.jp>
;;  All rights reserved.
;;
;;  $Id: uim-candidate.el,v 1.25 2005/10/21 19:09:18 nosuke Exp $
;;
;;  Redistribution and use in source and binary forms, with or
;;  without modification, are permitted provided that the
;;  following conditions are met:
;;
;;  1. Redistributions of source code must retain the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer.
;;  2. Redistributions in binary form must reproduce the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer in the documentation and/or other
;;     materials provided with the distribution.
;;  3. Neither the name of authors nor the names of its
;;     contributors may be used to endorse or promote products
;;     derived from this software without specific prior written
;;     permission.
;;
;;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
;;  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
;;  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
;;  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;;  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
;;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
;;  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
;;  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;;  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
;;  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;

;;
;; Checks that the candidate list can be displayed inlinely on current window.
;;
(defun uim-check-candidate-space ()
  ;; don't show the candidate in mini-buffer
  ;; also thin window is not supported
  (and (= (minibuffer-depth) 0)
       (>= (window-width) uim-candidate-minimum-width)))



;;
;; Format candlist
;;
(defun uim-make-candlist-line (cand index)
  ;; cand: ( <nil|t> "candstr" )

  (let ((candstr (cdr cand))
	(margin 1)
	appendix candlist)

    (when uim-candidate-display-frame
      (setq margin (+ margin 2)) ;; for | on both side
      )


    (let (tab)
      (when (setq tab (string-match "\t" candstr))
	(setq appendix (substring candstr (+ tab 1)))
	(setq candstr (substring candstr 0 tab))))

    ;; add index number and selected mark to candstr
    (if (car cand)
	(setq candstr (format "%2d.[%s]" index candstr))
      (setq candstr (format "%2d. %s " index candstr)))
    
    (if (< (uim-string-width candstr) 
	   (- (window-width) (- margin 1)))

	(setq candlist (list (list index 
				   candstr
				   (car cand) appendix)))

      (while (> (uim-string-width candstr) 0)
	;; split long candstr
	;;  maximum length is (window-width - 1)
	(let (trunc)
	  (setq trunc 
		(truncate-string-to-width candstr (- (window-width) margin)))

	  (setq candlist (append candlist 
				 (list (list index trunc (car cand) appendix))))

	  (setq candstr (substring candstr (length trunc)))

	  )))

    candlist
    ))

;;
;; Make candidate list to be displayed
;;
(defun uim-make-candlist (candidate) 
  ;; candidate: ( (cand# . candtotal) ( <nil|t> . "candstr") ... )

  (let ((cand candidate) 
	candnum
	candf)

    (setq candnum (list (list 0  (format "%d/%d" (caar cand) (cdar cand)))))
			  
    (setq cand (cdr cand))

    (let ((i 1))
      (while cand
	(nconc candnum (uim-make-candlist-line (car cand) i))
	(setq i (+ i 1))
	(setq cand (cdr cand))))
    candnum
    ))



;;
;; Put overlay
;;
(defun uim-set-candidate-overlay (index selected begin length)
  (let (face ol)
    (cond 
     (selected
      (setq face 'uim-candidate-selected-face))
     ((= index 0)
      (setq face 'uim-candidate-nth-face))
     ((= (% index 2) 1)
      (setq face 'uim-candidate-odd-face))
     ((= (% index 2) 0)
      (setq face 'uim-candidate-even-face)))

    (setq ol (make-overlay begin (+ begin length)))
    (overlay-put ol 'face face)
    (overlay-put ol 'priority 10)
    (setq uim-candidate-overlays
	  (cons ol uim-candidate-overlays))
    ))


;;
;; Merge formatted text into current buffer
;;
(defun uim-insert-candidate (candf)
  ;; candf: ( (0 string) (index string selected appendix) ...)
  (let (maxwidth offset ol col)

    ;; get max width candidate
    (let ((c candf) (max 0) len)
      (while c
	(when (stringp (car (cdar c)))
	  (setq len (uim-string-width (car (cdar c))))
	  (if (> len max) (setq max len)))
	(setq c (cdr c)))
      (setq maxwidth max))


    (let ((page (car (cdar candf)))
	  pagef)
      (setq pagef
	    (if uim-candidate-display-frame
		(concat (make-string (- maxwidth (uim-string-width page)) ?-)
			page)
	      (uim-format-string page maxwidth t)))
      (setq candf (cdr candf))

      (nconc candf (list (list 0 pagef))))


    (goto-char uim-candidate-start)


    ;; make overlay to chase base point
    (setq ol (make-overlay (point) (+ (point) 1)))

    ;; save original string
    (setq uim-candidate-original-str
	  (buffer-substring uim-candidate-original-start
			    uim-candidate-original-end))


    (if (and uim-preedit-cursor 
	     (> uim-preedit-cursor (point-min)))
	(setq col (make-overlay (- uim-preedit-cursor 1)
				uim-preedit-cursor)))

    ;; set overlay for chasing
    (uim-preedit-position-to-overlay)

    ;; remove tabs/spaces
    (uim-tab-pad-space uim-candidate-original-start
		       uim-candidate-original-end)

    ;; set preedit faces before merging
    (uim-set-preedit-faces t)
 
    ;; update uim-candidate-start
    (let (base-ofs)
      (setq base-ofs (- (- (overlay-end ol) 1) uim-candidate-start))
      (setq uim-candidate-start (+ uim-candidate-start base-ofs)))

    (delete-overlay ol)
    (setq ol nil)

    (goto-char uim-candidate-start)

    (save-excursion
      (uim-vertical-motion 0)
      (setq offset 
	    (uim-string-width (buffer-substring (point) uim-candidate-start))))

    ;; if offset + maxwidth >= window-width then reduce offset
    (let ((candwidth maxwidth))
      (if uim-candidate-display-frame 
	  (setq candwidth (+ candwidth 2)))
      (if (>= (+ offset candwidth) (window-width))
	  (setq offset (- (- (window-width) 1) candwidth))))


    (save-excursion
      (let ((strl candf) candidx candstr candsel)

	(if uim-show-candidate-upward
	    (uim-vertical-motion (- (length candf))) 
	  (uim-vertical-motion uim-candidate-vofs))

	(let ((i 1) 
	      mergeline
	      vhead linetmp padding overflow)

	  (save-excursion
	    (setq mergeline (uim-vertical-motion (length strl))))

	  ;; overwrite existing line
	  (while (<= i mergeline)

	    (setq candidx (nth 0 (car strl)))
	    (setq candstr (uim-format-string (nth 1 (car strl)) maxwidth))
	    (setq candsel (nth 2 (car strl)))

	    (if uim-candidate-display-frame
		(if (= candidx 0)
		    (setq candstr (concat "+" candstr "+"))
		  (setq candstr (concat "|" candstr "|"))))

	    ;; go virtual head of the next line
	    (if (not uim-show-candidate-upward)
		(uim-vertical-motion 1))

	    (setq vhead (point))

	    ;; make new line
	    (save-excursion

	      (end-of-line)
    
	      ;; |ABCDEFGH\ + |___aaa__|
	      ;; |IJKLMN  |
	      ;;
	      ;; get region from vhead to physical end
	      ;;
	      ;; linetmp = ABCDEFGH
	      ;; 
	      (setq linetmp 
		    (truncate-string-to-width (buffer-substring vhead (point))
					      (window-width)))

	      ;; make padding by truncating linetmp to offset size
	      ;;  (add space as padding if shortage)
	      ;;   witdth of padding is equal to offset
	      ;; 
	      ;; padding = ABC
	      ;;
	      (setq padding
		    (truncate-string-to-width linetmp offset nil 32))

	      ;; after candidate part
	      ;; 
	      ;;  overflow = GH
	      ;;
	      (let ((candwidth maxwidth))

		(if uim-candidate-display-frame 
		    (setq candwidth (+ candwidth 2)))

		(if (>= (uim-string-width linetmp) (+ offset candwidth))
		  (setq overflow 
			(truncate-string-to-width linetmp 
						  (uim-string-width linetmp)
						  (+ offset candwidth) 32))
		  (setq overflow "")))

	    
	      ;; index string selected appendix		    
	      (save-excursion
		(goto-char vhead)
		(delete-region vhead (+ vhead (length linetmp)))

		(insert (concat padding 
				candstr
				overflow)))

	      (uim-set-candidate-overlay candidx candsel 
					 (+ vhead (length padding))
					 (length candstr))


	      ;; update merge-end
	      (save-excursion
		(end-of-line)
		(setq uim-candidate-end (point)))
	      
	      (setq strl (cdr strl))
	      (setq i (+ i 1))
	      )

	    (if uim-show-candidate-upward
		(uim-vertical-motion 1))

	    ))


	;; append 
	(let (vhead)

	  (while strl

	    (setq candidx (nth 0 (car strl)))
	    (setq candstr (uim-format-string (nth 1 (car strl)) maxwidth))
	    (setq candsel (nth 2 (car strl)))

	    (if uim-candidate-display-frame
		(if (= candidx 0)
		    (setq candstr (concat "+" candstr "+"))
		  (setq candstr (concat "|" candstr "|"))))

	    (goto-char (point-max))

	    (setq vhead (+ (point) 1))

	    (insert (concat "\n"
			    (make-string offset 32)
			    candstr))

	    (uim-set-candidate-overlay candidx candsel 
				       (+ vhead offset)
				       (length candstr))

	    (save-excursion
	      (end-of-line)
	      (setq uim-candidate-end (point)))

	    (setq strl (cdr strl))))
	))

    (force-mode-line-update)

    ;; update cursor position
    (when col
      (goto-char (overlay-end col))
      (setq uim-preedit-cursor (point))
      (delete-overlay col))

    ))



(defun uim-check-overlay (head tail)
  (if (> head (point-min))
      (setq head (- head 1)))

  (if (< tail (point-max))
      (setq tail (+ tail 1)))
  
  (overlays-in  head tail))


;;
;; Insert candidate string
;;
(defun uim-show-candidate (candidate)

  (let (candf rect)
    ;; append index number and split candidate list from uim-agent
    (setq candf (uim-make-candlist candidate))


    ;; check whether candidate can be displayed downward or not.
    ;; downward -> upward -> scroll and downward (-> echo region)
    (save-excursion
      (goto-char uim-candidate-start)

      (if uim-candidate-display-inline

	  (let ((winofs (uim-get-window-offset))
		ol-down dhead dtail uhead utail)
	    
	    (if (and (not (setq ol-down (uim-check-overlay
					 (save-excursion
					   (beginning-of-line)
					   (setq dhead (point)))
					 (save-excursion
					   (uim-vertical-motion 
					    (+ uim-candidate-vofs 
					       (length candf)))
					   (end-of-line)
					   (setq dtail(point))))))
		     (>= (- (- (window-height) 1) (+ winofs 1)) 
			 (length candf)))

		(progn
		  (setq uim-show-candidate-upward nil)
		  (setq uim-candidate-original-start dhead)
		  (setq uim-candidate-original-end dtail)
		  )

	      (if (and (>= winofs (length candf))
		       (not (uim-check-overlay
			     (save-excursion
			       (uim-vertical-motion (- (length candf)))
			       (beginning-of-line)
			       (setq uhead (point)))
			     (save-excursion
			       (uim-vertical-motion -1)
			       (end-of-line)
			       (setq utail (point)))))
		       )
		  (progn
		    (setq uim-show-candidate-upward t)
		    (setq uim-candidate-original-start uhead)
		    (setq uim-candidate-original-end utail)
		    )
		       
		;; scroll anyway
		(if (not ol-down)
		    (let* ((fspace (- (- (- (window-height) 1) winofs) 1))
			   (vshift (- (length candf) fspace)))
		      (setq uim-show-candidate-upward nil)
		      (setq uim-window-force-scrolled t)

		      (setq uim-candidate-original-start dhead)
		      (setq uim-candidate-original-end dtail)

		      (if (> vshift winofs)
			  (recenter 0)
			(save-excursion 
			  (uim-debug (format "vshift %s" vshift))
			  (uim-vertical-motion (- (- (window-height)
						     (+ (length candf)) 2)))
			  (recenter 0))))
		  ;; disable inline display
		  (setq uim-candidate-original-start nil)
		  (setq uim-candidate-original-end nil)

		  (uim-debug (format "ol-down %s" ol-down))
		  
		  ))))))

    

    (if (and uim-candidate-display-inline
	     uim-candidate-original-start)
	;; inline candidate display mode
	(let ((inhibit-read-only t))
	  (uim-insert-candidate candf))

      ;; display candidate in minibuffer 
      (let* (out
	     (cand nil)
	     (page nil)
	     (select-start 0)
	     (select-end 0)
	     )

	(while candf

	  (if (not page)
	      ;; 1st entry
	      (setq page (car (cdar candf)))

	    (if (and (= select-start 0)
		     (nth 2 (car candf)))
		(setq select-start (length cand)))

	    (setq cand (concat cand (car (cdar candf))))

	    (if (nth 2 (car candf))
		(setq select-end (length cand)))

	    )
	    
	  (setq candf (cdr candf)))
	

	(if (and (or uim-xemacs
		     (< emacs-major-version 21))  ;; xemacs or emacs-20.x
		 (>= (uim-string-width (concat page cand)) 
		     (- (window-width) 1)))

	    (let ((rspace (- (- (window-width) 1) (string-width page)))
		  (candwidth (string-width cand))
		  (selwidth (string-width (substring cand 0 select-end))))

	      (if (and (> rspace 0)
		       (< rspace candwidth))
		  (if (< rspace (+ (- candwidth selwidth) 4))
		      ;; very long
		      (setq out (concat page " ..."
					(truncate-string-to-width
					 (truncate-string-to-width cand
								   candwidth
								   (- selwidth 4))
					 (- rspace 7))
					"..."
					))

		    (setq out 
			  (concat page " ..."
				  (truncate-string-to-width cand
							    candwidth
							    (+ (- candwidth
							       rspace) 4)))))
		(setq out (concat page cand))))
	  (setq out (concat page cand)))
	(let (message-log-max)
	  (uim-debug (format "%s" out))
	  (message out))
	)
      (uim-preedit-position-to-overlay)
      (uim-set-preedit-faces)

      ))
  ;(setq uim-candidate-displayed t)
  )


;;
;; Remove candidate
;;
(defun uim-remove-candidate ()
  (if (and uim-candidate-display-inline
	   uim-candidate-original-start)
      (let ((inhibit-read-only t))

	;; delete region which includes candidate
	(delete-region uim-candidate-original-start 
		       uim-candidate-end)

	;; restore original region
	(save-excursion
	  (goto-char uim-candidate-original-start)
	  (insert uim-candidate-original-str))

	;; delete overlay
	(mapcar 'delete-overlay uim-candidate-overlays)
	(setq uim-candidate-overlays nil)

	)

    ;; clear minibuffer
    (let (message-log-max)
      (message nil))
   
    )
  
  )


(provide 'uim-candidate)
