;; 
;;  uim-preedit.el
;;
;;  Copyright (c) 2005, Konosuke Watanabe <nosuke@csc.ne.jp>
;;  All rights reserved.
;;
;;  $Id: uim-preedit.el,v 1.15 2005/10/21 19:09:18 nosuke Exp $
;;
;;  Redistribution and use in source and binary forms, with or
;;  without modification, are permitted provided that the
;;  following conditions are met:
;;
;;  1. Redistributions of source code must retain the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer.
;;  2. Redistributions in binary form must reproduce the above
;;     copyright notice, this list of conditions and the
;;     following disclaimer in the documentation and/or other
;;     materials provided with the distribution.
;;  3. Neither the name of authors nor the names of its
;;     contributors may be used to endorse or promote products
;;     derived from this software without specific prior written
;;     permission.
;;
;;  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
;;  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
;;  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
;;  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;;  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
;;  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;;  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
;;  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
;;  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
;;  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;;  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
;;  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
;;  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
;;

;;
;; Select suitable font-face for the preedit string.
;;
(defun uim-select-preedit-face (attr)
  (cond
   ((string-match "s" attr)
    'uim-separator-face)
   ((string-match "ru" attr)
    'uim-preedit-highlight-underline-face)
   ((string-match "r" attr)
    'uim-preedit-highlight-face)
   ((string-match "u" attr)
    'uim-preedit-underline-face)
   (t 
    'uim-preedit-face)))

;;
;; Function for chasing overlay of preedit string
;;
(defun uim-preedit-overlay-insert-head (ol after head tail &optional len)
  (if after
      (move-overlay ol (+ (overlay-start ol) (- tail head))
		    (overlay-end ol))))


;;
;; Insert preedit
;;
(defun uim-insert-preedit (preedit)
  ;; preedit: ( (rct 美しい) (ust |) (ut 日本語) )

  (when uim-preedit-display-fences
    (setq preedit (append (cons '(t "|") preedit) '((t "|")))))

  (setq uim-preedit-cursor nil)

  ;; disregard read-only temporarily
  (let ((inhibit-read-only t))

    (save-excursion
      (setq uim-candidate-start nil)
      (setq uim-candidate-vofs 0)
      
      (let ((preedit-blocks preedit) block-start-point ol)

	(setq uim-preedit-start (point))


	(while preedit-blocks

	  (let (cand-start-cand
		(preedit-flag (format "%s" (caar preedit-blocks)))
		(preedit-str (car (cdar preedit-blocks))))

	    ;; save point for candidate displaying before insertion
	    ;;  i.e. head of the block
	    (if (string-match "c" preedit-flag)
	    (progn
	      (if (> (length preedit-str) 0)
		  (setq uim-candidate-start (point))
		;; workaround for uim-prime 
		(setq uim-candidate-start block-start-point))))

	    (setq block-start-point (point))

	    (when (> (length preedit-str) 0)
	      (insert preedit-str)

	      (setq uim-preedit-position-list 
		    (cons (list block-start-point
				(- (point) 1)
				(uim-select-preedit-face preedit-flag))
			  uim-preedit-position-list))
	      )


	    ;; save point for cursor displaying after insertion
	    (if (string-match "c" preedit-flag)
	    (setq uim-preedit-cursor (point)))

	    ;; update preedit-end
	    (setq uim-preedit-end (point))

	    (setq preedit-blocks (cdr preedit-blocks)))))

      )


    ;; "c" flag couldn't be found in preedit 
    (if (not uim-candidate-start)
	(setq uim-candidate-start uim-preedit-start))

    ;; set vertical offset
    (setq uim-candidate-vofs 
	  (uim-vertical-distance uim-candidate-start uim-preedit-cursor))
    )

  (setq uim-buffer-read-only buffer-read-only)
  (setq buffer-read-only t)

  )


(defun uim-preedit-position-to-overlay ()
  (mapcar '(lambda (x)
	     (let* ((head (nth 0 x))
		    (tail (nth 1 x))
		    (face (nth 2 x))
		    (headol (make-overlay head (+ head 1)))
		    (tailol (make-overlay tail (+ tail 1))))

	       (setq uim-preedit-ol-list 
		     (cons (list headol tailol face)
			   uim-preedit-ol-list))
	       (setq headol nil)
	       (setq tailol nil)
	       ))
	  uim-preedit-position-list)
  (setq uim-preedit-position-list nil)
  )


;; this function discard uim-preedit-ol-list

(defun uim-set-preedit-faces (&optional textprop-only)
  (let ((inhibit-read-only t))
    (let (ol phead ptail ol-face ol-info ol-head ol-tail)
      (while uim-preedit-ol-list
	(setq ol-info (car uim-preedit-ol-list))
	(setq ol-head (nth 0 ol-info))
	(setq ol-tail (nth 1 ol-info))
	(setq ol-face (nth 2 ol-info))
	(setq phead (- (overlay-end ol-head) 1))
	(setq ptail (overlay-end ol-tail))
    
	(put-text-property phead ptail 'face ol-face)
;;	(uim-debug (format "put-text-property %s %s %s" phead ptail ol-face))


	(when (and (not textprop-only)
		   (overlays-in phead ptail))
	  (setq ol (make-overlay phead ptail))
	  (overlay-put ol 'face ol-face)
	  (overlay-put ol 'priority 5)
	  (setq uim-preedit-overlays
		(cons ol uim-preedit-overlays))
	  (setq ol nil)
	  )

	(delete-overlay ol-head)
	(delete-overlay ol-tail)
	(setq uim-preedit-ol-list (cdr uim-preedit-ol-list))
	)
      )))



;; Remove preedit
(defun uim-remove-preedit ()

  (goto-char uim-preedit-start)

  (mapcar 'delete-overlay uim-preedit-overlays)
  (setq uim-preedit-overlays nil)

  (let ((inhibit-read-only t))

    (save-excursion
      ;; remove preedit string
      (delete-region uim-preedit-start uim-preedit-end))
  
    )

  (setq buffer-read-only uim-buffer-read-only)
  )


(provide 'uim-preedit)

